"""Rigid motion parameter generation routines."""

__all__ = ["rigid_motion"]


from typing import Sequence


import numpy as np
import numba as nb


def rigid_motion(
    ndim: int,
    nframes: int,
    degree: str | Sequence[float] = "moderate",
    seed: int = 42,
):
    """
    Generate rigid motion pattern as a Markov Chain process.

    Parameters
    ----------
    ndim : int
        Generate 2D (in-plane only) or 3D motion pattern.
    nframes : int
        Number of motion frames.
    degree : str | Sequence[float], optional
        Severity of motion. The default is ``"moderate"``.
    seed : int, optional
        Random number generator seed.
        The default is ``42``.

    Notes
    -----
    Severity of motion can be specified via the ``degree`` argument.
    This can be a string - accepted values are ``"subtle"``, ``"moderate"``
    and ``"severe"``. These corresponds to the following motion ranges:

    * ``"subtle"``: maximum rotation ``5.0 [deg]``; maximum translation ``2.0 [mm]``
    * ``"moderate"``: maximum rotation ``10.0 [deg]``; maximum translation ``8.0 [mm]``
    * ``"severe"``: maximum rotation ``16.0 [deg]``; maximum translation ``16.0 [mm]``

    As an alternative, user can specify a tuple of floats, where ``degree[0]``
    is the maximum rotation in ``[deg]`` and ``degree[1]`` is the maximum translation
    in ``[mm]``.

    Returns
    -------
    tuple[np.ndarray]
        Motion parameters. The number of outputs depend on the value of ``ndims``.

        For ``ndim == 2``:

        * angleZ : tnp.ndarray
            Rotation about ``z`` axis in ``[deg]`` of shape ``(nframes,)``.
        * dy : np.ndarray
            Translation towards ``y`` axis in ``[mm]`` of shape ``(nframes,)``.
        * dx : np.ndarray
            Translation towards ``x`` axis in ``[mm]`` of shape ``(nframes,)``.

        For ``ndim == 3``:

        * angleX : np.ndarray
            Rotation about ``x`` axis in ``[deg]`` of shape ``(nframes,)``.
        * angleY : torch.Tensor
            Rotation about ``y`` axis in ``[deg]`` of shape ``(nframes,)``.
        * angleZ : tnp.ndarray
            Rotation about ``z`` axis in ``[deg]`` of shape ``(nframes,)``.
        * dz : np.ndarray
            Translation towards ``z`` axis in ``[mm]`` of shape ``(nframes,)``.
        * dy : np.ndarray
            Translation towards ``y`` axis in ``[mm]`` of shape ``(nframes,)``.
        * dx : np.ndarray
            Translation towards ``x`` axis in ``[mm]`` of shape ``(nframes,)``.


    Example
    -------
    >>> from mrtwin import rigid_motion

    We can generate a 6-degree rigid motion pattern for 1000 motion states as:

    >>> roll, pitch, yaw, dz, dy, dx = rigid_motion(ndim=3, nframes=1000)

    """
    # Markov rate (I don't remember what this is :()
    rate = [[0.9, 0.05, 0.05], [0.4, 0.3, 0.3], [0.4, 0.3, 0.3]]
    transition_mtx = np.array(rate, np.float32)

    # generate probability array
    np.random.seed(seed)
    change = np.random.rand(6, nframes)

    # generate six random series
    x = _generate_series(6, nframes, transition_mtx, change)

    # rescale series
    x_max = np.abs(x).max(axis=1)[:, None]
    x_max[x_max == 0] = 1
    x = x / x_max
    x = x.astype(np.float32)

    # get motion range
    if isinstance(degree, str):
        if degree == "subtle":
            degree = [5.0, 2.0]
        elif degree == "moderate":
            degree = [10.0, 8.0]
        elif degree == "severe":
            degree = [16.0, 16.0]
        else:
            raise ValueError(
                f"Severity of motion not recognized - must be either 'subtle', 'moderate', 'severe' or a (rotation, translation) tuple in (deg, mm). Found {degree}."
            )

    # set
    roll = degree[0] * x[0]  # deg, rotation around x
    pitch = degree[0] * x[1]  # deg, rotation around y
    yaw = degree[0] * x[2]  # deg, rotation around z
    dx = degree[1] * x[3]  # mm, translation around x
    dy = degree[1] * x[4]  # mm, translation around x
    dz = degree[1] * x[5]  # mm, translation around x

    if ndim == 2:
        return yaw, dy, dx
    elif ndim == 3:
        return roll, pitch, yaw, dz, dy, dx
    else:
        raise ValueError(f"Invalid number of dims! must be 2 or 3 - found {ndim}")


# %% local utils
# adapted from
# https://ipython-books.github.io/131-simulating-a-discrete-time-markov-chain/

# The statespace
states = np.array([0, -1, 1], np.int64)


@nb.njit(fastmath=True, parallel=True)  # pragma: no cover
def _generate_series(n_parameters, n_frames, transition_matrix, change):
    # pre-allocate state history
    state_history = np.zeros((n_parameters, n_frames), dtype=np.int64)

    # initialize states
    current_state = np.zeros(n_parameters, dtype=np.int64)

    for p in nb.prange(n_parameters):
        for t in range(1, n_frames):
            current_state[p] = _generate_state(
                transition_matrix[current_state[p]], change[p, t]
            )
            state_history[p, t] = state_history[p, t - 1] + states[current_state[p]]

    return state_history


@nb.njit(fastmath=True, cache=True)  # pragma: no cover
def _generate_state(probability, change):
    if change <= probability[0]:
        out_state = 0
    elif change <= probability[0] + probability[1]:
        out_state = 1
    else:
        out_state = 2

    return out_state



"""Shape broadcasting sub-routines."""

__all__ = ["_expand_shapes"]

import numpy as np


def _expand_shapes(*shapes):  # noqa
    shapes = [list(shape) for shape in shapes]
    max_ndim = max(len(shape) for shape in shapes)

    shapes_exp = [np.asarray([1] * (max_ndim - len(shape)) + shape) for shape in shapes]
    shapes_exp = np.stack(shapes_exp, axis=0)  # (nshapes, max_ndim)
    shapes_exp = np.max(shapes_exp, axis=0)

    # restore original shape in non-padded portions
    shapes_exp = [list(shapes_exp[: -len(shape)]) + shape for shape in shapes]

    return tuple(shapes_exp)



"""Two-pool (Water + MT) BrainWeb phantom builder class."""

__all__ = [
    "FuzzyMTBrainwebPhantom",
    "CrispMTBrainwebPhantom",
    "NumericMTBrainwebPhantom",
]

import numpy as np

from .. import _classes

from ._brainweb import (
    FuzzyBrainwebPhantom,
    CrispBrainwebPhantom,
    NumericBrainwebPhantom,
)


class FuzzyMTBrainwebPhantom(FuzzyBrainwebPhantom):
    """Fuzzy MT BrainWeb phantom builder."""

    def get_model(self, B0: float):
        """Initialize model.

        Parameters
        ----------
        B0, float
            Static field strength in [T].

        """
        model_dict = _classes.tissue_map("mt-model")
        self._label = []
        self._properties = {"MVF": [], "T1w": [], "T2w": [], "k": []}
        for tissue in model_dict:
            self._label.append(tissue["Label"])
            self._properties["MVF"].append(tissue["MVF"])
            self._properties["T1w"].append(tissue["T1w"])
            self._properties["T2w"].append(tissue["T2w"])
            self._properties["k"].append(tissue["k"])

        # cast to array
        self._label = np.asarray(self._label, dtype=int)
        self._properties["MVF"] = np.asarray(self._properties["MVF"], dtype=np.float32)
        self._properties["T1w"] = np.asarray(self._properties["T1w"], dtype=np.float32)
        self._properties["T2w"] = np.asarray(self._properties["T2w"], dtype=np.float32)
        self._properties["k"] = np.asarray(self._properties["k"], dtype=np.float32)

    @property
    def MVF(self):  # noqa
        return self._properties["MVF"]

    @property
    def T1(self):  # noqa
        return self._properties["T1w"]

    @property
    def T2(self):  # noqa
        return self._properties["T2w"]

    @property
    def k(self):  # noqa
        return self._properties["k"]

    @property
    def properties(self):  # noqa
        _properties = {"weight": self.MVF, "T1": self.T1, "T2": self.T2, "k": self.k}
        return _properties


class CrispMTBrainwebPhantom(FuzzyMTBrainwebPhantom, CrispBrainwebPhantom):  # noqa
    """Crisp MT BrainWeb phantom builder."""

    pass


class NumericMTBrainwebPhantom(FuzzyMTBrainwebPhantom, NumericBrainwebPhantom):  # noqa
    """Numeric MT BrainWeb phantom builder."""

    pass



"""Test transmit field maps generation."""

import pytest
import numpy as np

from mrtwin import b1field


@pytest.mark.parametrize("shape", [(128, 128), (64, 64), (64, 64, 64)])
def test_b1field_basic_shape(shape):
    """
    Test the b1field function with basic input shapes.
    """
    # Call the b1field function
    b1map = b1field(shape)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Check that the output has the correct shape (for default nmodes=1)
    expected_shape = shape
    assert (
        b1map.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {b1map.shape}."


@pytest.mark.parametrize("nmodes", [2, 4])
def test_b1field_nmodes(nmodes):
    """
    Test the b1field function with different number of modes.
    """
    shape = (128, 128)

    # Call the b1field function with different nmodes
    b1map = b1field(shape, nmodes=nmodes)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Check that the output has the correct shape based on nmodes
    expected_shape = (nmodes,) + shape
    assert (
        b1map.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {b1map.shape}."


@pytest.mark.parametrize("b1range", [(0.5, 2.0), (0.8, 1.2), (1.0, 3.0)])
def test_b1field_b1range(b1range):
    """
    Test the b1field function with different b1range values.
    """
    shape = (128, 128)

    # Call the b1field function with different b1range
    b1map = b1field(shape, b1range=b1range)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate that the output shape is correct
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."

    # Check that values are within the specified b1range
    assert np.all(b1map >= b1range[0]) and np.all(b1map <= b1range[1]), (
        f"B1 map values should be in range {b1range}, but got min {b1map.min()} "
        f"and max {b1map.max()}"
    )


@pytest.mark.parametrize("dphi", [0.0, 45.0, 90.0])
def test_b1field_dphi(dphi):
    """
    Test the b1field function with different coil rotation angles (dphi).
    """
    shape = (128, 128)

    # Call the b1field function with different dphi values
    b1map = b1field(shape, dphi=dphi)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."


def test_b1field_shift():
    """
    Test the b1field function with a coil center shift.
    """
    shape = (128, 128)

    # Call the b1field function with a shift
    shift = (-3, 5)
    b1map = b1field(shape, shift=shift)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."


def test_b1field_with_mask():
    """
    Test the b1field function with a mask.
    """
    shape = (128, 128)

    # Create a binary mask
    mask = np.ones(shape)

    # Call the b1field function with a mask
    b1map = b1field(shape, mask=mask)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the shape of the output matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."


@pytest.mark.parametrize("coil_width", [1.0, 1.5, 2.0])
def test_b1field_coil_width(coil_width):
    """
    Test the b1field function with different coil widths.
    """
    shape = (128, 128)

    # Call the b1field function with different coil widths
    b1map = b1field(shape, coil_width=coil_width)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."


@pytest.mark.parametrize("ncoils", [2, 4, 8])
def test_b1field_ncoils(ncoils):
    """
    Test the b1field function with different numbers of coils.
    """
    shape = (128, 128)

    # Call the b1field function with different ncoils
    b1map = b1field(shape, ncoils=ncoils)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."


def test_b1field_with_3d_input():
    """
    Test the b1field function with 3D input data.
    """
    shape = (64, 64, 64)

    # Call the b1field function with 3D input
    b1map = b1field(shape)

    # Validate the output is a numpy ndarray
    assert isinstance(b1map, np.ndarray), "B1 map output should be a numpy ndarray."

    # Validate the output shape matches the input shape
    assert b1map.shape == shape, f"Expected shape {(1, *shape)}, but got {b1map.shape}."



"""Default path handling."""

__all__ = ["get_mrtwin_dir"]

import os

from pathlib import Path

from ._typing import CacheDirType


# Directory where data will be stored
def get_mrtwin_dir(cache_dir: CacheDirType = None) -> Path:
    """Get the MRTWIN directory.

    Adapted from brainweb_dl._brainweb.get_brainweb_dir

    Parameters
    ----------
    osf_dir : os.PathLike
       osf_directory to download the data.

    Returns
    -------
    os.PathLike
        Path to osf_dir

    Notes
    -----
    The osf directory is set in the following order:
    - The cache_dir passed as argument.
    - The environment variable MRTWIN_DIR.
    - The default osf_directory ~/.cache/cache_dir.
    """
    if cache_dir is not None:
        cache_dir = Path(cache_dir)
    elif "MRTWIN_DIR" in os.environ:
        cache_dir = Path(os.environ["MRTWIN_DIR"])
    else:
        cache_dir = Path.home() / ".cache" / "mrtwin"
    os.makedirs(Path(cache_dir), exist_ok=True)
    return Path(cache_dir)



"""OSF download tools."""

__all__ = ["get_osf_maps"]

import os

from pathlib import Path
from typing import Sequence

import numpy as np
import nibabel as nib

from osfclient import OSF

from .._utils import ssl_verification, CacheDirType

from .. import _prescription

# PREDATOR dataset project ID
DATASET_ID = "qkbca"

# +fmt: off
SUB_ID = [1, 2, 3, 4, 6, 8, 9, 10, 11, 12, 13, 14, 15, 17, 19, 22, 23, 25, 27, 28]
# +fmt: on


# Directory where data will be stored
def get_osf_dir(osf_dir: CacheDirType = None) -> Path:
    """Get the OSF directory.

    Adapted from brainweb_dl._brainweb.get_brainweb_dir

    Parameters
    ----------
    osf_dir : CacheDirType
       osf_directory to download the data.

    Returns
    -------
    os.PathLike
        Path to osf_dir

    Notes
    -----
    The osf directory is set in the following order:
    - The osf_dir passed as argument.
    - The environment variable OSF_DIR.
    - The default osf_directory ~/.cache/osf.
    """
    if osf_dir is not None:
        osf_dir = Path(osf_dir)
    elif "OSF_DIR" in os.environ:
        osf_dir = Path(os.environ["OSF_DIR"])
    else:
        osf_dir = Path.home() / ".cache" / "osf"
    os.makedirs(Path(osf_dir), exist_ok=True)
    return Path(osf_dir)


# Function to download and o rganize data per subject
def get_osf_maps(
    ndim: int,
    subject: int,
    shape: int | Sequence[int] | None = None,
    output_res: float | Sequence[float] | None = None,
    osf_dir: CacheDirType = None,
    force: bool = False,
    verify: bool = True,
):
    """
    Get quantitative maps adaped from Open Science CBS Neuroimaging Repository.

    Base resolution if 0.4 mm isotropic. Each subject dataset include
    M0, T1, T2, T2* and Chi values (skull-stripped). Values are
    extrapolated to 3T.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions. If ndim == 2, use a single slice
        (central axial slice).
    subject : int
        Subject id to download.
    shape: int | Sequence[int] | None, optional
        Shape of the output data, the data will be interpolated to the given shape.
        If int, assume isotropic matrix. The default is None (original shape).
    output_res: float | Sequence[float] | None, optional
        Resolution of the output data, the data will be rescale to the given resolution.
        If scalar, assume isotropic resolution. The default is None (keep original 0.4mm iso).
    osf_dir : CacheDirType, optional
        osf_directory to download the data.
        The default is None (~/.cache/osf).
    force : bool, optional
        Force download even if the file already exists.
        The default is False.
    verify : bool, optional
        Enable SSL verification.
        DO NOT DISABLE (i.e., verify=False)IN PRODUCTION.
        The default is True.

    Returns
    -------
    np.ndarray.
        3D/4D stacked array of shape (ncontrasts, *shape).
        The order of contrasts is:

            1. Equilibrium magnetization (M0)
            2. T1 in [ms]
            3. T2 in [ms]
            4. T2* in [ms]
            5. Chi in [ppb]

    """
    assert ndim == 2 or ndim == 3, ValueError(
        f"Number of spatial dimensions (={ndim}) must be either 2 or 3."
    )
    assert subject in SUB_ID, ValueError(
        f"subject (={subject}) must be one of {SUB_ID}"
    )

    # Default params
    if shape is not None and np.isscalar(shape):
        shape = shape * np.ones(ndim, dtype=int)
    if shape is not None:
        assert len(shape) == ndim, ValueError(
            "If shape is not None, it must be either a scalar or a ndim-length sequence."
        )
    if output_res is not None and np.isscalar(output_res):
        output_res = output_res * np.ones(ndim, dtype=int)
    if output_res is not None:
        assert len(output_res) == ndim, ValueError(
            "If output_res is not None, it must be either or a scalar a ndim-length sequence."
        )

    # Original resolution (0.4 mm iso)
    orig_res = 0.4 * np.ones(ndim)

    # Get base directory
    base_dir = get_osf_dir(osf_dir)

    # Create a subject-specific directory
    sub_dir = os.path.join(base_dir, f"sub{subject:02d}")
    os.makedirs(sub_dir, exist_ok=True)

    # Actual download
    if verify is False:
        with ssl_verification(verify=False):
            data = _actual_download(sub_dir, subject, force)
    else:
        data = _actual_download(sub_dir, subject, force)

    # Stack maps
    data = np.stack(
        [data["PD"], data["qT1"], data["qT2"], data["qT2STAR"], data["QSM"]], axis=0
    )

    # Select single slice
    if ndim == 2:
        center = int(data.shape[-3] // 2)
        data = data[:, center, :, :]

    # Make sure it is contiguous
    data = np.ascontiguousarray(data)

    if shape is None and output_res is None:
        data = np.nan_to_num(data, posinf=0.0, neginf=0.0)
        return data.astype(np.float32)
    elif output_res is None:
        output_res = orig_res  # 0.4 mm iso

    # Set prescription
    data = _prescription.set_prescription(
        data, orig_res, data.shape[-ndim:], output_res, shape
    )

    # Clean-up
    data = np.nan_to_num(data, posinf=0.0, neginf=0.0)
    return data.astype(np.float32)


def _actual_download(sub_dir, subject, force):
    # Initialize OSF client
    osf = OSF()

    # Replace 'your_osf_project_id' with your actual OSF project ID
    project = osf.project(DATASET_ID)

    # Replace 'your_storage_name' with the actual storage name (e.g., 'osfstorage')
    storage = project.storage("osfstorage")

    for folder in storage.folders:
        # Check if file belongs to the subject (adjust according to your naming convention)
        if f"sub{subject:02d}" == folder.name:

            # initialize
            keys = []
            values = []

            for file in folder.files:
                file_path = os.path.join(sub_dir, file.name)

                # Erase file if force
                if os.path.exists(file_path) and force:
                    os.remove(file_path)

                # Skip( download if the file already exists
                if not os.path.exists(file_path):
                    with open(file_path, "wb") as f:
                        file.write_to(f)

                # Load data
                keys.append(_get_contrast(file.name))
                values.append(nib.load(file_path).get_fdata().T)

    return dict(zip(keys, values))


def _get_contrast(file_name):
    return file_name.split("_")[3]



"""Tissue model for different experiments."""

__all__ = ["tissue_map", "get_t1", "get_t2star"]

import os
import sys

import numpy as np

if sys.version_info > (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

from pathlib import Path

from brainweb_dl._brainweb import _load_tissue_map

BUILT_IN_MAPS = ["single-pool", "mt-model", "mw-model", "mwmt-model"]


class TissueMap:
    """Tissue map files for the brainweb dataset."""

    single: Path = files("mrtwin._classes") / "single_pool.csv"  # type: ignore
    mt: Path = files("mrtwin._classes") / "mt_model.csv"  # type: ignore
    mw: Path = files("mrtwin._classes") / "mw_model.csv"  # type: ignore
    mwmt: Path = files("mrtwin._classes") / "mwmt_model.csv"  # type: ignore


def tissue_map(path_or_dict: str | os.PathLike | dict) -> list[dict]:
    """
    Build dictionary of tissue properties.

    Parameters
    ----------
    path_or_dict : str | os.PathLike | dict
        This can be either a list of dictionaries
        each describing a tissue class (e.g., WM, GM),
        a path to a cvs file storing the tissue classes
        descriptions or a string selecting one of the built-in
        tissue models. For the latter case, valid entries are:

            * "single-pool": Single pool tissue model.
            * "mw-model": Myelin Water (MW) + Free Water (Intra-Extracellular, IEW)
            * "mt-model": Macromolecular pool + Free Water (IEW + MW)
            * "mwmt-model": Macromolecular pool + MW + IEW

    Notes
    -----
    Each configuration file  field / dict must at least contain
    "Tissue Type", "ID" and "Label" fields. The other can be
    designed by the user to accomodate different properties (e.g., Diffusion).

    Returns
    -------
    list[dict]
        List of dictionaries each describing a tissue class.

    """
    # parse properties
    if isinstance(path_or_dict, str) and path_or_dict in BUILT_IN_MAPS:
        if path_or_dict == "single-pool":
            tissue_dict = _load_tissue_map(TissueMap.single)
        if path_or_dict == "mt-model":
            tissue_dict = _load_tissue_map(TissueMap.mt)
        if path_or_dict == "mw-model":
            tissue_dict = _load_tissue_map(TissueMap.mw)
        if path_or_dict == "mwmt-model":
            tissue_dict = _load_tissue_map(TissueMap.mwmt)
    else:
        if isinstance(path_or_dict, list):
            tissue_dict = path_or_dict
        else:
            tissue_dict = _load_tissue_map(path_or_dict)

    # iterate and cast string to float / int
    for item in tissue_dict:
        # check validity of dictionary
        assert "Tissue Type" in item and "ID" in item and "Label" in item, KeyError(
            "Tissue Type, ID and Label fields must be defined."
        )

        item["ID"] = int(item["Label"])
        for key in item.keys() - {"Tissue Type", "ID", "Label"}:
            item[key] = float(item[key])
            if item[key] == -1:
                item[key] = np.nan

    return tissue_dict


def get_t1(tissue: dict, B0: float, B0start: float) -> float:
    """
    Calculate / extrapolate T1 for a given tissue at a specific field strength.

    Parameters
    ----------
    tissue : dict
        Dictionary containing either tabulated T1
        or T1 model parameters (A, C).
    B0 : float
        Static field strength in [T].
    B0start, float
        Static field strength corresponding to tabulated T1.

    Returns
    -------
    float
        T1 value in [ms].

    """
    if np.isnan(tissue["T1"]):
        return model_t1(tissue["A"], tissue["C"], B0)
    return extrapolate_t1(tissue["T1"], B0start, B0)


def get_t2star(tissue: dict, B0: float, B0start: float) -> float:
    """
    Calculate / extrapolate T2* for a given tissue at a specific field strength.

    Parameters
    ----------
    tissue : dict
        Dictionary containing either tabulated T2*
        or T1 model parameters (T2, Chi).
    B0 : float
        Static field strength in [T].
    B0start, float
        Static field strength corresponding to tabulated T2*.

    Returns
    -------
    float
        T2 value in [ms].

    """
    if "Chi" in tissue:
        return model_t2star(tissue["T2"], tissue["Chi"], B0)
    return extrapolate_t2star(tissue["T2"], tissue["T2STAR"], B0start, B0)


def model_t1(A: float, C: float, B0: float) -> float:
    """
    Calculate T1 for a given tissue at a specific field strength.

    Parameters
    ----------
    A : float
        Base value in [ms].
    C : float
        Growth order.
    B0 : float
        Static field strength in [T].

    Returns
    -------
    float
        T1 value in [ms].

    """
    return A * (B0**C)


def model_t2star(T2: float, Chi: float, B0: float) -> float:
    """
    Calculate T2* for a given tissue at a specific field strength.

    Parameters
    ----------
    T2 : float
        Transverse relaxation time in [ms].
    Chi : float
        Magnetic susceptibility.
    B0 : float
        Static field strength in [T].

    Returns
    -------
    float
        T2* value in [ms].

    """
    gamma0 = 267.52219  # 10^6 rad⋅s−1⋅T⋅−1
    if T2 != 0:
        return 1 / (1 / T2 + gamma0 * np.abs(B0 * Chi))
    else:
        return 0.0


def extrapolate_t1(T1start: float, B0start: float, B0end: float) -> float:
    """
    Extrapolate T1 values at a given field strength.

    Parameters
    ----------
    T1start : float
        Initial T1 in [ms].
    B0start : float
        Initial field strength in [T].
    B0end : float
        Desired field strength in [T].

    Returns
    -------
    float
        Final T1 in [ms].

    """
    if B0start == B0end:
        return T1start
    scale = (B0end / B0start) ** 0.5
    return scale * T1start


def extrapolate_t2star(
    T2s_start: float, T2: float, B0start: float, B0end: float
) -> float:
    """
    Extrapolate T2* values at a given field strength.

    Parameters
    ----------
    T2s_start : float
        Initial T2* in [ms].
    T2 : float
        Transverse relaxation in [ms].
    B0start : float
        Initial field strength in [T].
    B0end : float
        Desired field strength in [T].

    Returns
    -------
    float
        Final T2* in [ms].

    """
    if B0start == B0end:
        return T2s_start
    scale = B0end / B0start
    R2 = 1 / (T2 + 1e-9)
    R2s_start = 1 / (T2s_start + 1e-9)
    R2p_start = R2s_start - R2
    R2p_start = np.clip(R2p_start, a_min=0.0, a_max=None)
    R2p_end = scale * R2p_start
    R2s_end = R2 + R2p_end
    T2s_end = 1 / (R2s_end + 1e-9)
    return T2s_end



"""Test GIRF generation."""

import pytest
import numpy as np

from mrtwin import generate_girf


@pytest.mark.parametrize(
    "dt, N",
    [
        (10e-6, 25000),
        (4e-6, 50000),
    ],
)
@pytest.mark.parametrize("fwhm", [5.0e3, [5.0e3, 10.0e3, 15.0e3]])
@pytest.mark.parametrize("delay", [0.1e-6, [0.1e-6, 0.2e-6, 0.3e-6]])
def test_generate_girf_basic(dt, N, fwhm, delay):
    """
    Test the generate_girf function for basic input values with both scalar and sequence delays/FWHMs.
    """
    freq_axis, girf = generate_girf(dt, N, fwhm, delay)

    # Check if the frequency axis and GIRF have the expected shapes
    assert isinstance(freq_axis, np.ndarray), "Frequency axis should be a NumPy array."
    assert isinstance(girf, np.ndarray), "GIRF should be a NumPy array."

    # Validate shapes
    assert freq_axis.shape == (
        N,
    ), f"Expected frequency axis shape {(N,)}, but got {freq_axis.shape}."
    assert girf.shape == (3, N), f"Expected GIRF shape (3, {N}), but got {girf.shape}."


@pytest.mark.parametrize("fwhm", [(5.0e3, 10.0e3, 15.0e3), 7.0e3])
@pytest.mark.parametrize("delay", [(0.1e-6, 0.2e-6, 0.3e-6), 0.2e-6])
def test_generate_girf_axis_independence(fwhm, delay):
    """
    Test the GIRF to ensure that each axis responds independently when provided with different FWHM and delay values.
    """
    if not (isinstance(fwhm, tuple)) and not (isinstance(delay, tuple)):
        pytest.skip("either fwhm or delay must be a tuple")

    dt = 4e-6
    N = 25000

    freq_axis, girf = generate_girf(dt, N, fwhm, delay)

    # Ensure that the GIRF contains different values for different axes when FWHM or delay varies
    assert not np.allclose(
        girf[0], girf[1]
    ), "GIRF for axis 0 and 1 should not be identical when parameters differ."
    assert not np.allclose(
        girf[1], girf[2]
    ), "GIRF for axis 1 and 2 should not be identical when parameters differ."
    assert not np.allclose(
        girf[0], girf[2]
    ), "GIRF for axis 0 and 2 should not be identical when parameters differ."


def test_generate_girf_single_axis_behavior():
    """
    Test the function with the same FWHM and delay for all axes and ensure uniform GIRF.
    """
    dt = 4e-6
    N = 25000
    fwhm = 7.0e3
    delay = 0.2e-6

    _, girf = generate_girf(dt, N, fwhm, delay)

    # Ensure that all axes should behave identically when FWHM and delay are the same
    assert np.allclose(
        girf[0], girf[1]
    ), "GIRF for all axes should be identical when FWHM and delay are identical."
    assert np.allclose(
        girf[1], girf[2]
    ), "GIRF for all axes should be identical when FWHM and delay are identical."



"""Shepp-Logan Phantom sub-package."""

__all__ = ["shepplogan_phantom"]

from typing import Sequence
from .._utils import CacheDirType, PhantomType

from ._shepplogan import NumericSheppLoganPhantom, CrispSheppLoganPhantom
from ._shepplogan_mw import NumericMWSheppLoganPhantom, CrispMWSheppLoganPhantom
from ._shepplogan_mt import NumericMTSheppLoganPhantom, CrispMTSheppLoganPhantom
from ._shepplogan_mwmt import NumericMWMTSheppLoganPhantom, CrispMWMTSheppLoganPhantom

VALID_MODELS = ["single-pool", "mt-model", "mw-model", "mwmt-model"]
VALID_SEGMENTATION = ["crisp"]


def shepplogan_phantom(
    ndim: int,
    shape: int | Sequence[int],
    model: str = "single-pool",
    segtype: str | bool = "crisp",
    B0: float = 1.5,
    cache: bool = None,
    cache_dir: CacheDirType = None,
) -> PhantomType:
    """
    Get SheppLogan phantom.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions. If ndim == 2, use a single slice
        (central axial slice).
    shape: int | Sequence[int]
        Shape of the output data. If int, assume isotropic matrix.
    model : str, optional
        String selecting one of the built-in
        tissue models. Valid entries are:

        * ``"single-pool"``: Single pool tissue model.
        * ``"mw-model"``: Myelin Water (MW) + Free Water (Intra-Extracellular, IEW)
        * ``"mt-model"``: Macromolecular pool + Free Water (IEW + MW)
        * ``"mwmt-model"``: Macromolecular pool + MW + IEW

        The default is ``"single-pool"``.
    segtype : str | bool, optional
        Phantom type. If it is a string (``"crisp"``)
        select crisp segmentation.
        If it is ``False``, return a dense numeric phantom.
        The default is ``crisp``.
    B0 : float, optional
        Static field strength in [T].
        The default is `1.5`.
    cache : bool | None, optional
        If ``True``, cache the phantom.
        The default is ``True`` for 3D phantoms
        and ``False`` for single-slice 2D.
    cache_dir : CacheDirType, optional
        cache_directory for phantom caching.
        The default is ``None`` (``~/.cache/mrtwin``).

    Returns
    -------
    PhantomType
        Shepp-Logan phantom.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from mrtwin import shepplogan_phantom

    We can generate a dense single slice 2D phantom with a matrix size
    of ``(200, 200)`` as:

    >>> phantom = shepplogan_phantom(ndim=2, shape=200, segtype=False)

    Phantom T1 and T2 maps, can be accessed as:

    >>> fig, ax = plt.subplots(2, 1)

    >>> im1 = ax[0].imshow(phantom.T1, cmap="magma", vmax=1500)
    >>> ax[0].axis("off"), ax[0].set_title("T1 [ms]")
    >>> fig.colorbar(im1, ax=ax[0], fraction=0.046, pad=0.04)

    >>> im2 = ax[1].imshow(phantom.T2, cmap="viridis", vmax=150)
    >>> ax[1].axis("off"), ax[1].set_title("T2 [ms]")
    >>> fig.colorbar(im, ax=ax[1], fraction=0.046, pad=0.04)

    """
    # check validity
    assert model in VALID_MODELS, ValueError(f"model must be one of {VALID_MODELS}")
    assert not (segtype) or segtype in VALID_SEGMENTATION, ValueError(
        f"segtype must be either False or one of {VALID_SEGMENTATION}"
    )

    # initialize model
    params = {
        "ndim": ndim,
        "shape": shape,
        "B0": B0,
        "cache": cache,
        "cache_dir": cache_dir,
    }
    if model == "single-pool":
        if segtype == "crisp":
            return CrispSheppLoganPhantom(**params)
        if segtype is False:
            return NumericSheppLoganPhantom(**params)
    if model == "mw-model":
        if segtype == "crisp":
            return CrispMWSheppLoganPhantom(**params)
        if segtype is False:
            return NumericMWSheppLoganPhantom(**params)
    if model == "mt-model":
        if segtype == "crisp":
            return CrispMTSheppLoganPhantom(**params)
        if segtype is False:
            return NumericMTSheppLoganPhantom(**params)
    if model == "mwmt-model":
        if segtype == "crisp":
            return CrispMWMTSheppLoganPhantom(**params)
        if segtype is False:
            return NumericMWMTSheppLoganPhantom(**params)



"""
===========================
Coil Sensitivity Simulation
===========================

Example of coil sensitivity map generation.

This examples show how to generate coil sensitivity maps
for a multi-channel receiver array.
"""

import matplotlib.pyplot as plt
import numpy as np

from mrtwin import sensmap

plt.rcParams["image.cmap"] = "gray"


# %%
# Basic Usage
# ===========
#
# Two- and three-dimensional coil sensitivity maps
# can be generated providing ``(nc, ny, nx)`` and
# ``(nc, nz, ny, nx)`` shaped tuple as a ``shape`` argument
# to `sensmap` routine, respectively:

smap2D = sensmap(shape=(8, 200, 200))  # 8 channels; (200, 200) matrix
print(smap2D.shape)

smap3D = sensmap(shape=(8, 128, 128, 128))  # 8 channels; (128, 128, 128) matrix
print(smap3D.shape)

display_magn = np.concatenate([np.abs(smap) for smap in smap2D], axis=1)
display_phase = np.concatenate([np.angle(smap) for smap in smap2D], axis=1)

fig1, ax1 = plt.subplots(2, 1)
ax1[0].imshow(display_magn), ax1[0].axis("off"), ax1[0].set_title("coil magnitudes")
ax1[1].imshow(display_phase), ax1[1].axis("off"), ax1[1].set_title("coil phases")
plt.show()

# %%
# Advanced Options
# ================
#
# The sensitivity maps can be altered by modifying several parameters:
#
# 1. ``coil_width``: width of the coil (with respect to FOV).
# 2. ``shift``: displacement of the center (in units of voxels).
# 3. ``dphi``: bulk rotation of the coil (in [deg]).
# 4. ``nrings``: number of rings for a cylindrical hardware setup.
#
# Without loss of generality, we show examples for 2D sensitivities:

# %%
# Coil width
# ----------
widths = [0.5, 1.0, 1.5, 2.0]
smap2D = [
    sensmap(shape=(8, 200, 200), coil_width=w)[0] for w in widths
]  # only show first channel

display = np.concatenate(smap2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.title(
    "coil width from 0.5 to 2.0 times fov (first channel)"
)
plt.show()

# %%
# Center shift
# ------------
dx = [-20, -10, 0, 10, 20]
smap2D = [
    sensmap(shape=(8, 200, 200), shift=(0, x), coil_width=0.5)[0] for x in dx
]  # for 3D, it would be shift=(dz, dy, dx)

display = np.concatenate(smap2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.title(
    "x-displacement from -20 to 20 times voxels (first channel)"
)
plt.show()

# %%
# Rotation
# --------
phi = [-20, -10, 0, 10, 20]
smap2D = [sensmap(shape=(8, 200, 200), dphi=angle, coil_width=0.5)[0] for angle in phi]

display = np.concatenate(smap2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.title(
    "coil rotation from -20 to 20 degrees (first channel)"
)
plt.show()

# %%
# Number of rings
# ---------------
nrings = [2, 4, 6, 8, 10]
smap2D = [sensmap(shape=(8, 200, 200), nrings=n, coil_width=0.5)[0] for n in nrings]

display = np.concatenate(smap2D, axis=1)

plt.figure()
plt.imshow(abs(display)), plt.axis("off"), plt.title(
    "number of rings from 2 to 10 (first channel)"
)
plt.show()


# %%
# Caching mechanism
# =================
#
# To reduce loading times, ``mrtwin`` implements a caching mechanism.
#
# If ``cache`` argument is set to ``True`` (default behaviour for ``ndim=3``), each sensitivity map
# segmentation (identified by the number of channels,
# matrix size, shift, rotation angle and number of rings)
# is saved on the disk in `npy` format.
#
# The path is selected according to the following hierachy (inspired by ``brainweb-dl``):
#
# 1. User-specific argument (``cache_dir``)
# 2. ``MRTWIN_DIR`` environment variable
# 3. ``~/.cache/mrtwin`` folder
#



"""Utils for file download."""

__all__ = ["ssl_verification"]

import requests
from contextlib import contextmanager


@contextmanager
def ssl_verification(verify=True):  # noqa
    # Default behaviour (do not disable)
    if verify:
        yield

    # Store the original `requests.Session.send` method
    original_send = requests.Session.send

    # Define a new `send` method that disables SSL verification
    def send_with_ssl_disabled(self, *args, **kwargs):
        kwargs["verify"] = False
        return original_send(self, *args, **kwargs)

    # Replace the original `send` method with the new one
    requests.Session.send = send_with_ssl_disabled

    try:
        # Execute the code within the context
        yield
    finally:
        # Restore the original `send` method
        requests.Session.send = original_send
